import csv
import os
import random
import tkinter as Tk
import tkinter.font as tkFont
from tkinter import messagebox
from picoh import picoh

global phraseList, rowList, numberOfRows

speechFile = 'picohData/PicohSpeech.csv'
#directory = picoh.dir

numberOfRows = 0
phraseList = []
rowList = []

file = speechFile


class Phrase(object):

    def __init__(self, set, variable, text):
        self.set = set
        self.variable = variable
        self.text = text


def generateRow(phrase, rowNo, frame):
    row = []

    # print(phrase.text)
    e = Tk.Entry(frame, font=customFont)
    e.insert(0, phrase.set)
    e.config(width=5)
    e.bind("<Return>", callback)
    e.bind("<FocusOut>", callback)
    e.grid(column=0, row=rowNo)
    row.append(e)

    e1 = Tk.Entry(frame, font=customFont)
    e1.insert(0, phrase.variable)
    e1.config(width=8)
    e1.bind("<Return>", callback)
    e1.bind("<FocusOut>", callback)
    e1.grid(column=1, row=rowNo)

    row.append(e1)

    e2 = Tk.Entry(frame, font=customFont)
    e2.insert(0, phrase.text)
    e2.config(width=89)
    e2.bind("<Return>", callback)
    e2.bind("<FocusOut>", callback)
    e2.grid(column=2, row=rowNo)

    e2.bind('<Control-a>', selAll)
    e2.bind('<Control-a>', selAll)

    row.append(e2)

    return row


def refreshCanvas():
    canvas.delete("all")

    canvas.create_window(0, 0, anchor='n', window=frame)

    canvas.update_idletasks()

    canvas.configure(scrollregion=canvas.bbox('all'),
                     yscrollcommand=scroll_y.set)

    rowHeight = 26
    blockMax = 400
    offset = blockMax - (rowHeight * numberOfRows)

    if numberOfRows * rowHeight < blockMax:
        frame.config(width=750, height=26 * numberOfRows, bg='white', highlightthickness=0)

        canvas.config(width=750, height=26 * numberOfRows, bg='white', highlightthickness=0)

        canvasSpacer.config(width=750, height=offset, bg='white', highlightthickness=0)

    else:
        frame.config(width=750, height=400, bg='white', highlightthickness=0)

        canvas.config(width=750, height=400, bg='white', highlightthickness=0)

        canvasSpacer.config(width=750, height=0, bg='white', highlightthickness=0)


def selectall(event):
    event.widget.tag_add("sel", "1.0", "end")
    return "break"


def selAll(event):
    event.widget.select_range(0, len(event.widget.get()))
    return "break"


def new():
    global numberOfRows, phraseList, rowList

    newPhrase = Phrase(0, 0, "")
    row = generateRow(newPhrase, numberOfRows + 1, frame)
    phraseList.append(newPhrase)
    rowList.append(row)
    numberOfRows = numberOfRows + 1
    refreshCanvas()
    canvas.yview_moveto(1)


def callback(*args):
    with open(file, 'w') as writeFile:
        writer = csv.writer(writeFile, delimiter=',')
        writer.writerow(["Set", "Variable", "Phrase"])

        for row in rowList:
            set = row[0].get()

            variable = row[1].get()

            phrase = row[2].get()

            if phrase != '':
                writer.writerow([set, variable, phrase])


def on_closing():
    if messagebox.askokcancel("Quit", "Do you want to quit?"):
        callback()
        root.destroy()

with open(file, 'r')as f:

    data = csv.reader(f)

    for row in data:
        if row[2] != '' and row[0] != 'Set':
            if row[0] == '' and row[1] == '':
                newPhrase = Phrase('', '', row[2])
                phraseList.append(newPhrase)

            elif row[0] == '' and row[1] != '':
                newPhrase = Phrase('', int(row[1]), row[2])
                phraseList.append(newPhrase)

            elif row[0] != '' and row[1] == '':
                newPhrase = Phrase(row[0], '', row[2])
                phraseList.append(newPhrase)

            else:
                newPhrase = Phrase(row[0], row[1], row[2])
                phraseList.append(newPhrase)

root = Tk.Tk()
root.title("Picoh - Speech Database")
root.geometry('776x470')

root.configure(bg='white')
root.resizable(0, 0)

root.protocol("WM_DELETE_WINDOW", on_closing)

customFont = tkFont.Font(root, family="Letter Gothic Std", size=12)

canvasThree = Tk.Canvas(root)
canvas = Tk.Canvas(root)
canvasTwo = Tk.Canvas(root)
canvasSpacer = Tk.Canvas(root)

canvasTwo.config(width=750, height=75, bg='white')
canvasThree.config(width=750, height=25, bg='white')

# canvasTwo.geometry('200x150')
scroll_y = Tk.Scrollbar(root, orient="vertical", command=canvas.yview)

frame = Tk.Frame(canvas)
frameTwo = Tk.Frame(canvasTwo)
frameThree = Tk.Frame(canvasThree)

frame.bind_class("Text", "<Control-a>", selectall)

frameTwo.config(width=750, height=75, bg='white')
frameThree.config(width=750, height=25, bg='white')

# group of widgets

for phrase in phraseList:
    row = generateRow(phrase, numberOfRows, frame)
    numberOfRows = numberOfRows + 1
    rowList.append(row)

rowHeight = 26
blockMax = 400
offset = blockMax - (rowHeight * numberOfRows)

if numberOfRows * rowHeight < blockMax:
    frame.config(width=750, height=26 * numberOfRows, bg='white', highlightthickness=0)
    canvas.config(width=750, height=26 * numberOfRows, bg='white', highlightthickness=0)
    canvasSpacer.config(width=750, height=offset, bg='white', highlightthickness=0)

else:
    frame.config(width=750, height=400, bg='white', highlightthickness=0)
    canvas.config(width=750, height=400, bg='white', highlightthickness=0)
    canvasSpacer.config(width=750, height=0, bg='white', highlightthickness=0)

# put the frame in the canvas
canvas.create_window(0, 0, anchor='nw', window=frame)
canvasTwo.create_window(0, 0, anchor='nw', window=frameTwo)
canvasThree.create_window(0, 0, anchor='nw', window=frameThree)

# make sure everything is displayed before configuring the scrollregion
canvas.update_idletasks()

canvas.configure(scrollregion=canvas.bbox('all'),
                 yscrollcommand=scroll_y.set)

canvas.grid(column=0, row=1)
scroll_y.grid(column=1, row=1, sticky="ns")

canvasTwo.grid(column=0, row=3)
canvasSpacer.grid(column=0, row=2)
canvasThree.grid(column=0, row=0)

saveButton = Tk.Button(frameTwo, text="Save", command=callback, font=customFont)
#saveButton.grid(column=1, row=1, sticky='w')

plusImage = Tk.PhotoImage(file="./Images/plus.gif")
addButton = Tk.Button(frameTwo, image=plusImage, command=new, font=customFont)
addButton.grid(column=0, row=1, sticky='w')

picohImage = Tk.PhotoImage(file="./Images/picohlogoSmall.gif")
picohLabel = Tk.Label(frameTwo, image=picohImage)
picohLabel.grid(column=2, row=1, sticky='w')

setLabel = Tk.Label(frameThree, text="Set", font=customFont, width=5)
setLabel.grid(column=0, row=0, sticky='w')

variableLabel = Tk.Label(frameThree, text="Variable", font=customFont, width=10)
variableLabel.grid(column=1, row=0, sticky='w')

phraseLabel = Tk.Label(frameThree, text="Phrase (Leave blank to delete)", font=customFont)
phraseLabel.grid(column=2, row=0, sticky='w')

# Launch the GUI
root.mainloop()


def getPhrase(set, variable):
    possiblePhrases = []
    for phrase in phraseList:
        if phrase.set == set and phrase.variable == variable:
            possiblePhrases.append(phrase.text)
    length = len(possiblePhrases)
    if length == 0:
        print("No matching phrase found for set: " + str(set) + " variable: " + str(variable) + " in: " + speechFile)
        return ""
    elif length == 1:
        return possiblePhrases[0]
    else:
        return possiblePhrases[random.randint[0, length]]



import tkinter as Tk
import tkinter.font as tkFont
from tkinter import ttk
# from tkinter import *
# from tkinter import messagebox
from tkinter import OptionMenu
import os.path
# from os import path
import numpy as np
# from PIL import Image, ImageFilter
# import math
from lxml import etree
import os
from picoh import picoh
from copy import deepcopy

from threading import Timer

# Class to hold Eyeshape information. Same fields as Picoh.obe xml file.

class EyeShape(object):

    def __init__(self, name_value, hexString_value, autoMirror_value, pupilRangeX_value, pupilRangeY_value):
        self.name = name_value
        self.hexString = hexString_value
        self.autoMirror = autoMirror_value
        self.pupilRangeX = pupilRangeX_value
        self.pupilRangeY = pupilRangeY_value


class PicohEyeDesigner(Tk.Frame):
    # Class variables

    # %%%%
    picohConnected = picoh.connected

    # Setup variables.
    clickedDown = False
    pupilActive = True
    drawing = False
    startedMoving = False

    currentfilename = ""

    # Binary grids, one for each button.
    gridArray = np.zeros((9, 8))
    gridArrayOne = np.zeros((9, 8))
    gridArrayTwo = np.zeros((9, 8))
    gridArrayThree = np.zeros((9, 8))
    gridArrayFour = np.zeros((9, 8))
    gridArrayFive = np.zeros((9, 8))

    buttonArray = []
    buttonArrayOne = []
    buttonArrayTwo = []
    buttonArrayThree = []
    buttonArrayFour = []
    buttonArrayFive = []

    # List of EyeShape objects.
    shapeList = []

    # Coordinates for top left of window
    rootx = 20
    rooty = 40

    # Variables to hold colour and size preferences.
    bgCol = 'white'
    textCol = 'black'
    buttonCol = 'white'
    # pupilButtonHighlightColour = '#408bf9'
    pupilButtonHighlightColour = 'SkyBlue1'

    buttonWidth = 10
    tickWidth = 15

    tree = None



    def initialize(self):

        # Configure Window

        self.parent.title("Picoh Eye Shape Designer")
        self.parent.grid_rowconfigure(1, weight=0)
        self.parent.grid_columnconfigure(1, weight=0)

        self.customFont = tkFont.Font(family="Letter Gothic Std", size=12)

        self.frame = Tk.Frame(self.parent)
        self.frame.configure(bg=self.bgCol)

        self.screen_width = root.winfo_screenwidth()
        self.screen_height = root.winfo_screenheight()

        #Variables to track tick boxes:
        self.pupilVar = Tk.BooleanVar()
        self.pupilVar.set(True)

        self.pupilTrack = Tk.BooleanVar()
        self.pupilTrack.set(False)

        self.mirrorVar = Tk.IntVar()
        self.mirrorVar.set(0)

        self.speak = Tk.IntVar()
        self.speak.set(0)

        self.rangeVar = Tk.IntVar()
        self.rangeVar.set(0)

        # Create popups for rename and new shape.

        self.entryPopTwo = Tk.Entry(self.frame, width=30, text="Test", font=self.customFont)
        self.entryPopTwo.bind('<Return>', self.rename)

        self.entryPop = Tk.Entry(self.frame, width=30, text="Test", font=self.customFont)
        self.entryPop.bind('<Return>', self.newShape)

        # Add pupil overlay  and track checkboxes

        checkbox = Tk.Checkbutton(self.frame, text="Overlay Pupil", variable=self.pupilVar, command=self.checkBoxAction)
        checkbox.grid(row=4, rowspan=1, column=18, columnspan=5, sticky="w")
        checkbox.configure(bg=self.bgCol, font=self.customFont)

        pupilTrackBox = Tk.Checkbutton(self.frame, text="Mouse-Pupil", variable=self.pupilTrack,
                                       command=self.pupilTrackAction)
        pupilTrackBox.grid(row=8, rowspan=1, column=27, columnspan=4, sticky="w")
        pupilTrackBox.configure(bg=self.bgCol, font=self.customFont, width=self.tickWidth)

        # Labels

        l1 = Tk.Label(self.frame, text="Eyeshape")
        l1.grid(row=0, column=0, columnspan=4, sticky="W", padx=(10, 0))
        l1.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        l2 = Tk.Label(self.frame, text="Pupil")
        l2.grid(row=0, column=9, columnspan=3, sticky="W")
        l2.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        l3 = Tk.Label(self.frame, text="Blink 1")
        l3.grid(row=10, column=0, columnspan=4, sticky="W", padx=(10, 0))
        l3.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        l4 = Tk.Label(self.frame, text="Blink 2")
        l4.grid(row=10, column=9, columnspan=3, sticky="W")
        l4.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        l5 = Tk.Label(self.frame, text="Blink 3")
        l5.grid(row=10, column=18, columnspan=3, sticky="W")
        l5.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        l6 = Tk.Label(self.frame, text="Blink 4")
        l6.grid(row=10, column=27, columnspan=3, sticky="W")
        l6.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        self.textLab = Tk.Label(self.frame, text='Are You Sure?', font=self.customFont)

        self.filenamelabel = Tk.Label(self.frame, text="")
        #            self.filenamelabel.grid(row=13,column = 0,columnspan = 10,sticky = "W", padx = (10,0))

        #  Create 2D arrays with 0's to hold button states.
        for x in range(0, 6):
            for j in range(9):
                column = []
                for i in range(8):
                    column.append(0)
                self.getButtonArray(x).append(column)

        # New Button
        self.newButton = Tk.Button(self.frame, text="New", command=self.newButton, width=self.buttonWidth)
        self.newButton.grid(row=4, column=27, columnspan=4, sticky="w")
        self.newButton.configure(highlightbackground=self.bgCol, font=self.customFont)

        # Rename Button
        self.renameButton = Tk.Button(self.frame, text="Rename", command=self.renameButton)
        self.renameButton.grid(row=4, column=31, columnspan=4, sticky="e")
        self.renameButton.configure(highlightbackground=self.bgCol, font=self.customFont, width=self.buttonWidth)

        # Duplicate Button
        self.dupButton = Tk.Button(self.frame, text="Duplicate", command=self.duplicate, width=self.buttonWidth)
        self.dupButton.grid(row=5, column=31, columnspan=4, sticky="e")
        self.dupButton.configure(highlightbackground=self.bgCol, font=self.customFont)

        # Delete button
        self.delButton = Tk.Button(self.frame, text="Delete", command=self.deleteShapeButton, width=self.buttonWidth)
        self.delButton.grid(row=5, column=27, columnspan=4, sticky="w")
        self.delButton.configure(highlightbackground=self.bgCol, font=self.customFont)

        # Test blink button
        self.blinkButton = Tk.Button(self.frame, text="Test Blink", command=self.testBlink, width=self.buttonWidth)
        self.blinkButton.grid(row=7, column=31, columnspan=4, sticky="e")
        self.blinkButton.configure(highlightbackground=self.bgCol, font=self.customFont, width=9)

        # Speak tick box, have picoh read out file names and changes.
        self.speakTickBox = Tk.Checkbutton(self.frame, text="Speak", variable=self.speak)
        # self.speakTickBox.grid(row=8, column=31, columnspan = 4, sticky="e")
        self.speakTickBox.config(bg=self.bgCol, highlightcolor=self.textCol, font=self.customFont, width=self.tickWidth)

        #  Reset buttons for each grid

        self.resetButton = Tk.Button(self.frame, text='Clear', command=lambda: self.reset(0))
        self.resetButton.grid(row=0, column=5, columnspan=3, sticky="E")
        self.resetButton.configure(highlightbackground=self.bgCol, bg=self.bgCol, fg=self.textCol, font=self.customFont)

        self.resetButtonOne = Tk.Button(self.frame, text="Clear", command=lambda: self.reset(1))
        self.resetButtonOne.grid(row=0, column=14, columnspan=3, sticky="E")
        self.resetButtonOne.configure(highlightbackground=self.bgCol, font=self.customFont)

        self.resetButtonTwo = Tk.Button(self.frame, text="Clear", command=lambda: self.reset(2))
        self.resetButtonTwo.grid(row=10, column=5, columnspan=3, sticky="E")
        self.resetButtonTwo.configure(highlightbackground=self.bgCol, font=self.customFont)

        self.resetButtonThree = Tk.Button(self.frame, text="Clear", command=lambda: self.reset(3))
        self.resetButtonThree.grid(row=10, column=14, columnspan=3, sticky="E")
        self.resetButtonThree.configure(highlightbackground=self.bgCol, font=self.customFont)

        self.resetButtonFour = Tk.Button(self.frame, text="Clear", command=lambda: self.reset(4))
        self.resetButtonFour.grid(row=10, column=23, columnspan=3, sticky="E")
        self.resetButtonFour.configure(highlightbackground=self.bgCol, font=self.customFont)

        self.resetButtonFive = Tk.Button(self.frame, text="Clear", command=lambda: self.reset(5))
        self.resetButtonFive.grid(row=10, column=32, columnspan=3, sticky="E")
        self.resetButtonFive.configure(highlightbackground=self.bgCol, font=self.customFont)

        # copy buttons
        copyDownButton = Tk.Button(self.frame, width=0, height=0, borderwidth=0,
                                   highlightthickness=-2, image=copyDown, padx=-2, pady=-2)

        copyDownButton.configure(highlightbackground=self.bgCol)

        copyRightOneButton = Tk.Button(self.frame, width=0, height=0, borderwidth=0,
                                       highlightthickness=-2, image=copyRight, padx=-2, pady=-2)

        copyRightOneButton.configure(highlightbackground=self.bgCol)

        copyRightTwoButton = Tk.Button(self.frame, width=0, height=0, borderwidth=0,
                                       highlightthickness=-2, image=copyRight, padx=-2, pady=-2)

        copyRightTwoButton.configure(highlightbackground=self.bgCol)

        copyRightThreeButton = Tk.Button(self.frame, width=0, height=0, borderwidth=0,
                                         highlightthickness=-2, image=copyRight, padx=-2, pady=-2)

        copyRightThreeButton.configure(highlightbackground=self.bgCol)

        # Buttons used during renaming or the creation of a new shape.
        self.but = Tk.Button(self.frame, text="Yes",
                             command=self.deleteShape, font=self.customFont, width=self.buttonWidth)

        self.butCancel = Tk.Button(self.frame, text="No",
                                   command=self.cancel, font=self.customFont, width=self.buttonWidth)

        self.okayOne = Tk.Button(self.frame, text="Okay", highlightbackground=self.bgCol,
                                 command=self.newShape, font=self.customFont, width=self.buttonWidth)

        self.cancelOne = Tk.Button(self.frame, text="Cancel", highlightbackground=self.bgCol,
                                   command=self.cancel, font=self.customFont, width=self.buttonWidth)

        self.okayTwo = Tk.Button(self.frame, text="Okay", highlightbackground=self.bgCol,
                                 command=self.rename, font=self.customFont, width=self.buttonWidth)

        self.cancelTwo = Tk.Button(self.frame, text="Cancel",
                                   command=self.cancel, highlightbackground=self.bgCol, font=self.customFont,
                                   width=self.buttonWidth)

        # Add copy buttons to grid.

        copyDownButton.grid(row=10, column=3)
        copyRightOneButton.grid(row=15, column=8)
        copyRightTwoButton.grid(row=15, column=17)
        copyRightThreeButton.grid(row=15, column=26)

        # Bind commands to copy buttons.
        copyDownButton.bind("<Button>", lambda event, grid=0: self.copyGrid(event, grid, grid + 2))
        copyDownButton.bind("<ButtonRelease-1>", self.OnMouseUp)
        copyRightOneButton.bind("<Button>", lambda event, grid=2: self.copyGrid(event, grid, grid + 1))
        copyRightOneButton.bind("<ButtonRelease-1>", self.OnMouseUp)
        copyRightTwoButton.bind("<Button>", lambda event, grid=3: self.copyGrid(event, grid, grid + 1))
        copyRightTwoButton.bind("<ButtonRelease-1>", self.OnMouseUp)
        copyRightThreeButton.bind("<Button>", lambda event, grid=4: self.copyGrid(event, grid, grid + 1))
        copyRightThreeButton.bind("<ButtonRelease-1>", self.OnMouseUp)

        # Picoh button, toggles sending data to Picoh. If not Picoh detected default to off.
        if self.picohConnected:
            chosenLogo = logoOn
        else:
            chosenLogo = logo

        # Create Picoh logo button.
        self.picohButton = Tk.Button(self.frame, command=self.picohToggle, image=chosenLogo)
        self.picohButton.grid(row=0, column=27, columnspan=20, rowspan=3, sticky="s")
        self.picohButton.configure(highlightbackground=self.bgCol)

        picohPanel = Tk.Label(self.frame, image=picohGraphic)
        #  picohPanel.grid(row=9, column=8, columnspan=16, rowspan=16, sticky="sw")

        #  Generate button grids: (xStart,yStart,grid)
        self.generateButtons(0, 1, 0)
        self.generateButtons(9, 1, 1)
        self.generateButtons(0, 11, 2)
        self.generateButtons(9, 11, 3)
        self.generateButtons(18, 11, 4)
        self.generateButtons(27, 11, 5)

        # Create a Tkinter variable

        self.tkvar = Tk.StringVar(self.frame)

        # Read in data from Picoh.obe xml file.
        self.xmlReadin()

        # Load the Shapelist with
        self.refreshShapeList()

        # Trace tkvar to enable shape chosen in drop down to be loaded
        self.tkvar.trace_id = self.tkvar.trace_variable("w", self.loadShape)

        self.saved = True

        # x and y range entry boxes

        self.xRangeVar = Tk.StringVar()
        self.xRangeVar.set('5')

        self.yRangeVar = Tk.StringVar()
        self.yRangeVar.set('5')

        self.xRangeEntry = Tk.Entry(self.frame, width=2, textvariable=self.xRangeVar)

        self.xRangeEntry.grid(row=7, column=22, columnspan=5, sticky='w')
        self.xRangeEntry.config(bg='white', font=self.customFont)

        self.yRangeEntry = Tk.Entry(self.frame, width=2, textvariable=self.yRangeVar)

        self.yRangeEntry.grid(row=8, column=22, columnspan=5, sticky='w')
        self.yRangeEntry.config(bg='white', font=self.customFont)

        self.xRangeLabel = Tk.Label(self.frame, text="Pupil Range X", font=self.customFont)

        self.xRangeLabel.grid(row=7, column=18, columnspan=5, sticky='w')
        self.xRangeLabel.config(bg=self.bgCol, fg=self.textCol)

        self.yRangeLabel = Tk.Label(self.frame, text="Pupil Range Y")

        self.yRangeLabel.grid(row=8, column=18, columnspan=5, sticky='w')
        self.yRangeLabel.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        self.xRangeVar.trace_variable("w", self.updateRange)
        self.yRangeVar.trace_variable("w", self.updateRange)

        # Create check boxes

        self.mirrorCheckbox = Tk.Checkbutton(self.frame, text="Auto Mirror", variable=self.mirrorVar,
                                             command=self.mirrorChange)
        self.mirrorCheckbox.grid(row=7, rowspan=1, column=27, columnspan=4, sticky="w")
        self.mirrorCheckbox.config(bg=self.bgCol, highlightcolor=self.textCol, font=self.customFont,
                                   width=self.tickWidth)

        self.rangeCheckbox = Tk.Checkbutton(self.frame, text="Show Pupil Range", variable=self.rangeVar,
                                            command=self.displayRange)
        self.rangeCheckbox.grid(row=5, rowspan=1, column=18, columnspan=5, sticky="w")
        self.rangeCheckbox.config(bg=self.bgCol, fg=self.textCol, font=self.customFont)

        # Pack frame.
        self.frame.pack(fill=Tk.X, padx=0, pady=0)

        root.bind('<Motion>', self.motion)

        #Load first shape in the list.
        self.shapeIndex = 0
        self.loadShape(True, shapeName=self.shapeList[self.shapeIndex].name, loading=True)
        self.updatePicoh()

        self.checkBoxAction()
        checkbox.invoke()

        if self.picohConnected:
            picoh.reset()
            picoh.close()

        root.mainloop()

    def __init__(self, parent):
        Tk.Frame.__init__(self, parent)
        self.parent = parent
        self.initialize()

    # Function to generate buttons
    def generateButtons(self, buttonStartX, buttonStartY, grid):
        for i in range(0, self.getGridArray(grid).shape[0]):
            for j in range(0, self.getGridArray(grid).shape[1]):

                b = Tk.Button(self.frame, highlightbackground=self.buttonCol, width=0, height=0, borderwidth=0,
                              highlightthickness=-2, padx=-2, pady=-2)

                if j == 0 and grid == 0 or j == 0 and grid == 2:
                    b.grid(row=i + buttonStartY, column=j + buttonStartX, padx=(10, 0))

                else:
                    b.grid(row=i + buttonStartY, column=j + buttonStartX)

                b.config(image=offImage)

                # Bind events
                b.bind("<B1-Motion>", lambda event, grid=grid: self.OnMouseMove(event, grid))
                b.bind("<Leave>", self.OnMouseLeave)
                b.bind("<Button>", lambda event, grid=grid, x=i, y=j: self.OnMouseDown(event, x, y, grid))
                b.bind("<ButtonRelease-1>", self.OnMouseUp)

                # Add button to button array
                self.getButtonArray(grid)[i][j] = b

    # Copies grid to destination.
    def copyGrid(self, event, grid, destination):

        if grid == 0:

            for i in range(0, 9):
                for j in range(0, 8):
                    if self.getGridArray(grid)[i][j]:
                        self.turnButtonOn(i, j, destination, loading=False)
                    else:
                        self.turnButtonOff(i, j, destination, loading=False)
        else:
            for i in range(0, 9):
                for j in range(0, 8):
                    if self.getGridArray(grid)[i][j]:
                        self.turnButtonOn(i, j, destination, loading=False)
                    else:
                        self.turnButtonOff(i, j, destination, loading=False)

        self.saved = False

    def removeFromXML(self, nameToDelete):

        root = self.tree.getroot()
        for channel in root:
            for item in channel:
                if item[0].text == nameToDelete:
                    channel.remove(item)
                    self.writeToFile()
                    return

    def renameInXML(self, nameToChange, newName):

        root = self.tree.getroot()
        for channel in root:
            for item in channel:
                if item[0].text == nameToChange:
                    item[0].text = newName
                    self.writeToFile()
                    return

    def updateXML(self, eyeShape):
        root = self.tree.getroot()
        for channel in root:
            for item in channel:
                if item[0].text == eyeShape.name:
                    item[2].text = str(eyeShape.pupilRangeX)
                    item[3].text = str(eyeShape.pupilRangeY)
                    item[5].text = eyeShape.hexString
                    if eyeShape.autoMirror:
                        item[6].text = 'true'
                    else:
                        item[6].text = 'false'

                    self.writeToFile()
                    return
        return

    def addToXML(self, eyeShape):

        root = self.tree.getroot()
        for channel in root:
            for item in channel:
                if item[0].text == self.tkvar.get():
                    newElement = deepcopy(item)
                    newElement[0].text = eyeShape.name
                    newElement[2].text = str(eyeShape.pupilRangeX)
                    newElement[3].text = str(eyeShape.pupilRangeY)
                    newElement[5].text = eyeShape.hexString
                    if eyeShape.autoMirror:
                        newElement[6].text = 'true'
                    else:
                        newElement[6].text = 'false'

                    channel.append(newElement)
                    self.writeToFile()
                    # print(eyeShape.name)
                    return

    def writeToFile(self):

        my_tree = self.tree

        #directory = picoh.dir

        #file = os.path.join(directory, 'Ohbot.obe')
        file = 'picohData/Ohbot.obe'

        with open(file, 'wb') as f:
            f.write(etree.tostring(my_tree))
            f.close()

    # Shows highlighted blue squares on pupil grid to indicate range.
    def displayRange(self):

        xRange = self.shapeList[self.shapeIndex].pupilRangeX
        yRange = self.shapeList[self.shapeIndex].pupilRangeY

        if xRange > 8:
            xRange = 8

        if yRange > 9:
            yRange = 9

        if xRange < 0:
            xRange = 0

        if yRange < 0:
            yRange = 0

        yStart = (3 - int(yRange / 2))
        xStart = (3 - int(xRange / 2))

        if xStart < 0:
            xStart = 0
        if yStart < 0:
            yStart = 0

        for i in range(0, 9):
            for j in range(0, 8):

                if self.gridArrayOne[i][j]:
                    self.buttonArrayOne[i][j].config(highlightbackground='white')

                else:
                    self.buttonArrayOne[i][j].config(highlightbackground=self.buttonCol)

        for i in range(xStart, xStart + xRange):
            for j in range(yStart, yStart + yRange):
                if self.rangeVar.get():
                    self.getButtonArray(1)[j][i].config(highlightbackground=self.pupilButtonHighlightColour)

        self.xRangeVar.set(str(xRange))
        self.yRangeVar.set(str(yRange))

    def pupilTrackAction(self):
        if self.pupilTrack.get():
            return
        else:
            picoh.move(picoh.EYETURN, 5)
            picoh.move(picoh.EYETILT, 5)

    # Sets pupil range, called when value is changed in entry box.
    def updateRange(self, *args):

        if self.xRangeVar.get() == '' or self.yRangeVar.get() == '':
            return
        self.shapeList[self.shapeIndex].pupilRangeX = int(self.xRangeVar.get())
        self.shapeList[self.shapeIndex].pupilRangeY = int(self.yRangeVar.get())
        self.displayRange()
        self.updateXML(self.shapeList[self.shapeIndex])

    def refreshShapeList(self):

        self.choices = []
        for entry in self.shapeList:
            self.choices.append(entry.name)

        # Get first item in list of choices and set as default
        self.choices.sort()

        # first = next(iter(self.choices), None)
        # self.tkvar.set(self.shapeList[0].name)

        # update popup menu with names from shapelist.

        self.popupMenu = Tk.OptionMenu(self.frame, self.tkvar, *self.choices)

        self.popupMenu.grid(row=3, column=27, columnspan=15, sticky="w")

        self.popupMenu.configure(bg=self.bgCol, width=30, font=self.customFont)

    def duplicate(self):

        currentShape = self.shapeList[self.shapeIndex]
        newEyeShape = EyeShape("New", "", False, 5, 5)
        newEyeShape.autoMirror = currentShape.autoMirror
        newEyeShape.hexString = currentShape.hexString
        newEyeShape.name = currentShape.name + " (Copy)"
        newEyeShape.pupilRangeX = currentShape.pupilRangeX
        newEyeShape.pupilRangeY = currentShape.pupilRangeY

        self.shapeList.append(newEyeShape)

        self.addToXML(newEyeShape)
        if self.speak.get() == 1:
            picoh.say(currentShape.name + " Duplicated")
        self.loadShape(shapeName=currentShape.name + " (Copy)", internal=True)
        self.updatePicoh()
        self.popupMenu.destroy()
        self.refreshShapeList()
        self.cancelNewShape()

    # Function to read XML files
    def xmlReadin(self):

        directory = picoh.dir

        #file = os.path.join(directory, 'Ohbot.obe')
        file = 'picohData/Ohbot.obe'

        self.tree = etree.parse(file)

        index = 0

        for element in self.tree.iter():

            if element.tag == "Name":
                self.shapeList.append(EyeShape(str(element.text), "", False, 5, 5))

            if element.tag == "PupilRangeX":
                self.shapeList[index].pupilRangeX = int(element.text)

            if element.tag == "PupilRangeY":
                self.shapeList[index].pupilRangeY = int(element.text)

            if element.tag == "Hex":
                self.shapeList[index].hexString = element.text

            if element.tag == "AutoMirror":
                if element.text == "true":
                    self.shapeList[index].autoMirror = True

                else:
                    self.shapeList[index].autoMirror = False

                index = index + 1

    def openHex(self, hexString, loading=False):

        # Empty binary strings

        shapeBinary = ''
        pupilBinary = ''
        blinkOneBinary = ''
        blinkTwoBinary = ''
        blinkThreeBinary = ''
        blinkFourBinary = ''

        # Load into binary strings for each char

        for hexBit in hexString[:18]:
            shapeBinary = shapeBinary + self.hexToBin(hexBit)

        for hexBit in hexString[90:108]:
            pupilBinary = pupilBinary + self.hexToBin(hexBit)

        for hexBit in hexString[18:36]:
            blinkOneBinary = blinkOneBinary + self.hexToBin(hexBit)
        for hexBit in hexString[36:54]:
            blinkTwoBinary = blinkTwoBinary + self.hexToBin(hexBit)
        for hexBit in hexString[54:72]:
            blinkThreeBinary = blinkThreeBinary + self.hexToBin(hexBit)
        for hexBit in hexString[72:90]:
            blinkFourBinary = blinkFourBinary + self.hexToBin(hexBit)

        # Load Matrix with binary strings.

        self.loadMatrix(pupilBinary, 1, loading)
        self.loadMatrix(shapeBinary, 0, loading)
        self.loadMatrix(blinkOneBinary, 2, loading)
        self.loadMatrix(blinkTwoBinary, 3, loading)
        self.loadMatrix(blinkThreeBinary, 4, loading)
        self.loadMatrix(blinkFourBinary, 5, loading)

    # Function to flip the state of a button at given coordinate.
    def flipButton(self, i, j, grid):

        if grid == 1:
            if self.gridArrayOne[i][j] == 0:
                self.turnPupilOff(i, j)
            else:
                self.turnPupilOn(i, j)

        if self.getGridArray(grid)[i][j] == 0:
            self.turnButtonOn(i, j, grid, loading=False)
        else:
            self.turnButtonOff(i, j, grid, loading=False)

        if grid == 1:
            self.updateRange()

        self.updatePicoh()

        # Turn pupil on at coordinate i, j

    # Turn pupil on at coordinate i,j
    def turnPupilOn(self, i, j):

        if self.gridArray[i][j]:
            self.buttonArray[i][j].config(image=onImage)
            self.buttonArray[i][j].config(highlightbackground='white')

        else:
            self.buttonArray[i][j].config(image=offImage)
            self.buttonArray[i][j].config(highlightbackground=self.buttonCol)

    # Turn pupil off at coordinate i,j
    def turnPupilOff(self, i, j):

        if self.pupilVar:
            self.buttonArray[i][j].config(highlightbackground=self.pupilButtonHighlightColour)
            self.buttonArray[i][j].config(image=offImage)

    # Turn button on at coordinate i,j
    def turnButtonOn(self, i, j, grid, loading):

        # Prevent eye shape from turning on if pupil is on at this location

        if self.gridArrayOne[i, j] and self.pupilVar and grid == 0 and not loading:
            return

        self.getGridArray(grid)[i][j] = 1
        self.getButtonArray(grid)[i][j].config(highlightbackground='white', image=onImage)

        self.saved = False

    # Turn button off at coordinate i,j
    def turnButtonOff(self, i, j, grid, loading):

        # Prevent eye shape from turning off if pupil is on at this location

        if self.gridArrayOne[i, j] and self.pupilVar and grid == 0 and not loading:
            return

        self.getGridArray(grid)[i][j] = 0
        self.getButtonArray(grid)[i][j].config(highlightbackground=self.buttonCol, image=offImage)

        self.saved = False

    """
    Returns a hex string representing the current state of all grids
    9 pairs of hex bits for each grid. Order: Eye,Pupil,Blink1,Blink2,Blink3,Blink4
    """

    def hexFromGrids(self):

        # Create an empty binary string and read each grid into it.
        binaryStringIn = ''
        order = [0, 2, 3, 4, 5, 1]

        for grid in order:

            for i in range(0, self.getGridArray(grid).shape[0]):
                for j in range(0, self.getGridArray(grid).shape[1]):
                    binaryStringIn = str(binaryStringIn) + str(int(self.getGridArray(grid)[i][j]))
                    hd = (len(binaryStringIn) + 3) // 4

        string = '%.*x' % (hd, int('0b' + binaryStringIn, 0))
        return string

    # Get hex string from grids and set Picoh's eyes to it.
    def updatePicoh(self):
        if self.picohConnected:
            hexToSend = self.hexFromGrids()
            picoh._setEyes(hexToSend, hexToSend)

    # Toggle sending data to Picoh.
    def picohToggle(self):

        if self.picohConnected:
            self.picohButton.config(image=logo)
            self.picohConnected = False
        else:
            self.picohButton.config(image=logoOn)
            self.picohConnected = True
            self.updatePicoh()

    # Function refresh all grids

    def newButton(self):

        self.newButton.grid_remove()
        self.renameButton.grid_remove()

        self.dupButton.grid_remove()
        self.delButton.grid_remove()

        self.popupMenu.destroy()

        self.okayOne.grid(row=4, column=27, columnspan=4, sticky="w")
        self.cancelOne.grid(row=4, column=31, columnspan=4, sticky="e")
        self.entryPop.grid(row=3, column=27, columnspan=15, sticky="w")

    #  print(self.entryPop.get())

    def cancel(self):
        self.refreshShapeList()
        self.cancelNewShape()

    def cancelNewShape(self):
        self.entryPop.grid_remove()
        self.entryPopTwo.grid_remove()
        self.textLab.grid_remove()
        self.but.grid_remove()
        self.butCancel.grid_remove()
        self.okayOne.grid_remove()
        self.cancelOne.grid_remove()
        self.okayTwo.grid_remove()
        self.cancelTwo.grid_remove()

        self.popupMenu.grid(row=3, column=27, columnspan=15, sticky="w")
        self.newButton.grid(row=4, column=27, columnspan=4, sticky="w")
        self.renameButton.grid(row=4, column=31, columnspan=4, sticky="e")
        self.delButton.grid(row=5, column=27, columnspan=4, sticky="w")
        self.dupButton.grid(row=5, column=31, columnspan=4, sticky="e")

        self.entryPop.delete(0, Tk.END)
        self.entryPopTwo.delete(0, Tk.END)

    def newShape(self, *args):

        newName = self.entryPop.get()

        if newName == "":
            self.cancelNewShape()
            print("Please enter a name")
            return

        newEyeShape = EyeShape("New", "", False, 5, 5)
        newEyeShape.autoMirror = "False"
        newEyeShape.hexString = "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        newEyeShape.name = newName
        newEyeShape.pupilRangeX = 5
        newEyeShape.pupilRangeY = 5

        self.addToXML(newEyeShape)

        self.shapeList.append(newEyeShape)

        self.refreshShapeList()

        self.loadShape(shapeName=newName, internal=True)

        self.cancelNewShape()

        self.updatePicoh()

    def renameButton(self):

        self.newButton.grid_remove()
        self.renameButton.grid_remove()
        self.dupButton.grid_remove()
        self.delButton.grid_remove()
        self.popupMenu.destroy()

        self.entryPopTwo.grid(row=3, column=27, columnspan=10, sticky="w")
        self.okayTwo.grid(row=4, column=27, columnspan=4, sticky="w")
        self.cancelTwo.grid(row=4, column=31, columnspan=4, sticky="e")

        self.entryPopTwo.delete(0, Tk.END)
        self.entryPopTwo.insert(0, self.tkvar.get())

    def rename(self, *args):
        if self.entryPopTwo.get() == "":
            print("Please enter a name")
            self.cancelNewShape()
            return

        oldName = self.shapeList[self.shapeIndex].name
        newName = self.entryPopTwo.get()
        self.shapeList[self.shapeIndex].name = self.entryPopTwo.get()
        self.refreshShapeList()
        self.tkvar.set(self.entryPopTwo.get())
        self.cancelNewShape()
        self.renameInXML(oldName, newName)
        if self.speak.get() == 1:
            picoh.say(oldName + " renamed to " + newName)

    def deleteShapeButton(self):

        self.newButton.grid_remove()
        self.renameButton.grid_remove()
        self.dupButton.grid_remove()
        self.delButton.grid_remove()
        self.popupMenu.destroy()

        self.textLab.grid(row=3, column=27, columnspan=5, sticky="ws")
        self.textLab.config(bg=self.bgCol, font=self.customFont)

        self.but.grid(row=4, column=27, columnspan=4, sticky="w")
        self.but.configure(highlightbackground=self.bgCol, font=self.customFont)

        self.butCancel.grid(row=4, column=31, columnspan=4, sticky="e")
        self.butCancel.configure(highlightbackground=self.bgCol, font=self.customFont)

    def deleteShape(self):

        for idx, shape in enumerate(self.shapeList):
            if shape.name == self.shapeList[self.shapeIndex].name:
                self.removeFromXML(shape.name)
                del self.shapeList[idx]
                break
        self.refreshShapeList()
        self.loadShape(True, self.shapeList[0].name)
        self.cancelNewShape()
        # self.tkvar.set(self.shapeList[0].name)
        self.updatePicoh()

        if self.speak.get() == 1:
            picoh.say(shape.name + " deleted")

    # Load shape and set grids to it.
    def loadShape(self, internal, shapeName, loading=False, *args):

        if shapeName:
            for index, shape in enumerate(self.shapeList):
                if shape.name == shapeName:
                    chosenShape = shape
                    self.shapeIndex = index
        else:
            for index, shape in enumerate(self.shapeList):
                if shape.name == self.tkvar.get():
                    chosenShape = shape
                    self.shapeIndex = index

        # if loading == True:
        if self.speak.get() == 1:
            picoh.say(chosenShape.name + " loaded")

        self.openHex(chosenShape.hexString, loading)

        self.filenamelabel.config(text=chosenShape.name, font=self.customFont)
        self.parent.title("Picoh Eye Shape Designer - " + chosenShape.name)
        self.currentfilename = chosenShape.name

        self.displayRange()

        self.xRangeVar.set(str(chosenShape.pupilRangeX))
        self.yRangeVar.set(str(chosenShape.pupilRangeY))

        if chosenShape.autoMirror:
            self.mirrorVar.set(1)
        else:
            self.mirrorVar.set(0)

        picoh.move(picoh.EYETILT, 5)
        picoh.move(picoh.EYETURN, 5)
        self.updatePicoh()
        self.pupilTrackAction()

        self.tkvar.trace_vdelete("w", self.tkvar.trace_id)
        self.tkvar.set(chosenShape.name)

        self.tkvar.trace_id = self.tkvar.trace_variable("w", self.loadShape)

        self.checkBoxAction()
        self.checkBoxAction()

    # Check box action for automirror check box.
    def mirrorChange(self):

        if self.mirrorVar.get() == 1:
            self.shapeList[self.shapeIndex].autoMirror = True
        else:
            self.shapeList[self.shapeIndex].autoMirror = False
        self.updatePicoh()
        picoh.move(picoh.EYETURN, 5)
        picoh.move(picoh.EYETILT, 5)
        self.updateXML(self.shapeList[self.shapeIndex])

    # For a given hex bit return the binary string.
    @staticmethod
    def hexToBin(hexBit):

        if hexBit == '0':
            return "0000"
        elif hexBit == '1':
            return "0001"
        elif hexBit == '2':
            return "0010"
        elif hexBit == '3':
            return "0011"
        elif hexBit == '4':
            return "0100"
        elif hexBit == '5':
            return "0101"
        elif hexBit == '6':
            return "0110"
        elif hexBit == '7':
            return "0111"
        elif hexBit == '8':
            return "1000"
        elif hexBit == '9':
            return "1001"
        elif hexBit == 'a' or hexBit == 'A':
            return "1010"
        elif hexBit == 'b' or hexBit == 'B':
            return "1011"
        elif hexBit == 'c' or hexBit == 'C':
            return "1100"
        elif hexBit == 'd' or hexBit == 'D':
            return "1101"
        elif hexBit == 'e' or hexBit == 'E':
            return "1110"
        elif hexBit == 'f' or hexBit == 'F':
            return "1111"
        else:
            print("not a hex char:")
            print(hexBit)
            return '0000'

    # Reset a given grid, clearing all buttons.
    def reset(self, grid):
        for j in range(8):
            for i in range(9):
                self.turnButtonOff(i, j, grid, loading=False)
                if grid == 1:
                    self.turnPupilOn(i, j)
                if grid == 0:
                    self.gridArray[i][j] = 0
        if grid == 1:
            self.updateRange()
        self.updatePicoh()
        self.updateXML(self.shapeList[self.shapeIndex])

    # Load a given grid with a binary string.
    def loadMatrix(self, string, grid, loading=False):
        count = 0
        for char in string:

            y = count % 8
            x = int(count / 8)

            if grid == 1:

                if char == '1':
                    self.turnButtonOn(x, y, grid, loading)
                    self.turnPupilOff(x, y)
                else:
                    self.turnButtonOff(x, y, grid, loading)
                    self.turnPupilOn(x, y)
            else:
                if char == '1':
                    self.turnButtonOn(x, y, grid, loading)
                else:
                    self.turnButtonOff(x, y, grid, loading)
            count += 1

    # Return the gridArray Object for a given grid number
    def getGridArray(self, grid):
        if grid == 0:
            return self.gridArray
        if grid == 1:
            return self.gridArrayOne
        if grid == 2:
            return self.gridArrayTwo
        if grid == 3:
            return self.gridArrayThree
        if grid == 4:
            return self.gridArrayFour
        if grid == 5:
            return self.gridArrayFive

    # Return the buttonArray Object for a given grid number
    def getButtonArray(self, grid):
        if grid == 0:
            return self.buttonArray
        if grid == 1:
            return self.buttonArrayOne
        if grid == 2:
            return self.buttonArrayTwo
        if grid == 3:
            return self.buttonArrayThree
        if grid == 4:
            return self.buttonArrayFour
        if grid == 5:
            return self.buttonArrayFive

    # Action for mouse down
    def OnMouseDown(self, event, x, y, grid):

        # Hasn't left a square yet so started moving is False
        startedMoving = False

        # Decided if the stroke should draw or erase nodes
        if self.getGridArray(grid)[x, y]:
            self.drawing = False
        else:
            self.drawing = True

        # Flip the button that triggered the event
        self.flipButton(x, y, grid)

        # Send new grid to Picoh
        if grid == 0 or grid == 1:
            self.updatePicoh()

    #  Action for mouse move
    def OnMouseMove(self, event, grid):

        # If mouse is not outside original button do nothing.

        if not self.startedMoving:
            return

        # Calculate an offset to account for the window being moved.

        offsetx = root.winfo_x() - self.rootx
        offsety = root.winfo_y() - self.rooty

        # Map the pixel coordinate of the event to the corresponding grid coordinate.

        coordinateX = ((event.x_root - 26 - offsetx) / 28)
        coordinateY = ((event.y_root - 50 - offsety) / 32) - 1

        # print(str(coordinateX)+"\n"+str(coordinateY))

        if coordinateY > 9:
            coordinateY = coordinateY + 0.5
            coordinateY = coordinateY % 10
        if coordinateX > 8:
            coordinateX = coordinateX % 9

        # Constrain coordinates to between 0 - 8.
        if coordinateX > 8:
            return
        if coordinateX > 7:
            coordinateX = 7
        if coordinateX < 0:
            coordinateX = 0

        if coordinateY > 8:
            coordinateY = 8

        if coordinateY < 0:
            coordinateY = 0

        # If in drawing mode, turn on the button at the coordinate and update Picoh's matrix.
        #  Otherwise, turn off button at the coordinate and update Picoh's matrix.

        if self.drawing:
            self.turnButtonOn(int(coordinateY), int(coordinateX), grid, loading=False)
            if grid == 1:
                self.turnPupilOff(int(coordinateY), int(coordinateX))
        else:
            self.turnButtonOff(int(coordinateY), int(coordinateX), grid, loading=False)
            if grid == 1:
                self.turnPupilOn(int(coordinateY), int(coordinateX))

        self.updatePicoh()

    def motion(self, event):

        if not self.pupilTrack.get():
            return

        x, y = event.x, event.y

        x = root.winfo_pointerx() - root.winfo_rootx()
        y = root.winfo_pointery() - root.winfo_rooty()

        parentName = event.widget.winfo_parent()
        parent = event.widget._nametowidget(parentName)  # event.widget is your widget

        scaledX = x / parent.winfo_width()
        scaledX = scaledX * 10

        scaledY = y / parent.winfo_height()
        scaledY = scaledY * 10

        """
        x = root.winfo_pointerx()

        y = root.winfo_pointery()
        abs_coord_x = root.winfo_pointerx() - root.winfo_rootx()

        abs_coord_y = root.winfo_pointery() - root.winfo_rooty()

        scaledX = abs_coord_x/x
        scaledX = scaledX*10

        scaledY = abs_coord_y/y
        scaledY = scaledY*10
        
        """

        if scaledX < 10 or scaledX > 0 or scaledY < 10 or scaledY > 0:
            picoh.move(picoh.EYETURN, scaledX)
            picoh.move(picoh.EYETILT, 10 - scaledY)

    def saveShape(self):
        self.shapeList[self.shapeIndex].hexString = self.hexFromGrids()
        self.saved = True

        # self.updateRange()
        self.updateXML(self.shapeList[self.shapeIndex])

    #    print("saved!")
    # print(self.shapeList[self.shapeIndex].hexString)

    # Mouse up action for buttons.
    def OnMouseUp(self, event):
        self.updateRange()
        self.clickedDown = False

        # self.uRange()

        if not self.saved:
            t = Timer(0.3, self.saveShape)
            t.start()  # after 1 second save.

    def testBlink(self):
        for x in range(10, 0, -2):
            picoh.move(picoh.LIDBLINK, x)
            picoh.wait(0.04)

        for x in range(0, 10, 2):
            picoh.move(picoh.LIDBLINK, x)
            picoh.wait(0.04)

    # Mouse leave action for buttons.
    def OnMouseLeave(self, event):
        self.startedMoving = True

    # Pupil toggle check box callback.
    def checkBoxAction(self):
        if self.pupilVar:
            self.pupilVar = False
            for j in range(8):
                for i in range(9):
                    self.turnPupilOn(i, j)
        else:
            self.pupilVar = True
            for j in range(8):
                for i in range(9):
                    if self.gridArrayOne[i][j] == 1:
                        self.turnPupilOff(i, j)


if __name__ == "__main__":
    root = Tk.Tk()
    xDim = 20
    yDim = 40
    hDim = 606
    wDim = 1000
    root.geometry('%dx%d+%d+%d' % (wDim, hDim, xDim, yDim))
    root.configure(bg='white')
    root.resizable(0, 0)
    onImage = Tk.PhotoImage(file="./Images/onsmaller.gif")
    offImage = Tk.PhotoImage(file="./Images/offsmaller.gif")
    logo = Tk.PhotoImage(file="./Images/picohlogo.gif")
    logoOn = Tk.PhotoImage(file="./Images/picohlogoOn.gif")
    spacer = Tk.PhotoImage(file="./Images/onsmaller.gif")
    copyDown = Tk.PhotoImage(file="./Images/movedown.gif")
    copyRight = Tk.PhotoImage(file="./Images/moveright.gif")
    picohGraphic = Tk.PhotoImage(file="./Images/picohGraphic.gif")

    app = PicohEyeDesigner(root)
    root.mainloop()
    os.system('''/usr/bin/osascript -e 'tell app "Finder" to set frontmost of process "Python" to true' ''')



# This is a script to help control a Picoh Robot. www.ohbot.co.uk

import platform
import serial
import serial.tools.list_ports
import time
import threading
import os
import os.path
from os import path
import shutil
import sys
import wave
import subprocess
from lxml import etree
import random
import re
import csv

if platform.system() == "Windows":
    import winsound
    # For SAPI speech
    from comtypes.client import CreateObject
if platform.system() == "Darwin":
    from playsound import playsound
if platform.system() == "Linux":
    from playsound import playsound

# define constants for motors
HEADNOD = 0
HEADTURN = 1
EYETURN = 2
LIDBLINK = 3
TOPLIP = 4
BOTTOMLIP = 5
EYETILT = 6

# array to hold 
sensors = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

# define a module level variable for the serial port
port = ""

# define library version
version = "1.0"

global writing, voice, synthesizer, speechRate, connected, shapeList, phraseList, topLipFree

# flag to stop writing when writing for threading
writing = False

# flag to allow the library to run when not connected
connected = False

shapeList = []
phraseList = []

# Flag to track if top lip is below centre.
topLipFree = False

# global to set the params to speech synthesizer which control the voice
if platform.system() == "Windows":
    voice = ""
if platform.system() == "Darwin":
    voice = "Alex"
if platform.system() == "Linux":
    voice = ""

# global to set the speech speed.
speechRate = 170

# Global flag to use a synthesizer other than sapi.
# If it's not sapi then it needs to support -w parameter to write to file e.g. espeak or espeak-NG
if platform.system() == "Windows":
    synthesizer = "sapi"
if platform.system() == "Darwin":
    synthesizer = "say -o "
if platform.system() == "Linux":
    synthesizer = "festival"
print("Speech Synthesizer: " + synthesizer)

dirName = 'picohData'

try:
    # Create target Directory
    os.mkdir(dirName)
except FileExistsError:
    pass

# Variable to hold the location of the picoh library folder.
dir = os.path.dirname(os.path.abspath(__file__))

if not path.exists('picohData/PicohSpeech.csv'):
    shutil.copyfile(os.path.join(dir, 'PicohSpeech.csv'),'picohData/PicohSpeech.csv')

if not path.exists('picohData/ohbot.obe'):
    shutil.copyfile(os.path.join(dir, 'ohbot.obe'),'picohData/ohbot.obe')


# Variables to hold name of speech database and eyeshape files.

speechDatabaseFile = 'picohData/PicohSpeech.csv'
eyeShapeFile = 'picohData/ohbot.obe'
speechAudioFile = 'picohData/picohspeech.wav'


# Cache of pupil positions
global lastfex, lastfey
lastfex = 5
lastfey = 5

ser = None

class EyeShape(object):

    def __init__(self, name_value, hexString_value, autoMirror_value, pupilRangeX_value, pupilRangeY_value):
        self.name = name_value
        self.hexString = hexString_value
        self.autoMirror = autoMirror_value
        self.pupilRangeX = pupilRangeX_value
        self.pupilRangeY = pupilRangeY_value


class Phrase(object):

    def __init__(self, set, variable, text):
        self.set = set
        self.variable = variable
        self.text = text


# Read eyeshape file into eyeshape list.
def _loadEyeShapes():
    global shapeList

   # dir = os.path.dirname(os.path.abspath(__file__))

    #file = os.path.join(dir, eyeShapeFile)
    file = eyeShapeFile

    tree = etree.parse(file)

    index = 0

    for element in tree.iter():

        if element.tag == "Name":
            shapeList.append(EyeShape(str(element.text), "", False, 5, 5))

        if element.tag == "PupilRangeX":
            shapeList[index].pupilRangeX = int(element.text)

        if element.tag == "PupilRangeY":
            shapeList[index].pupilRangeY = int(element.text)

        if element.tag == "Hex":
            shapeList[index].hexString = element.text

        if element.tag == "AutoMirror":
            if element.text == "true":
                shapeList[index].autoMirror = True

            else:
                shapeList[index].autoMirror = False

            index = index + 1

# Read speech database file into phraseList.
def _loadSpeechDatabase():
    global phraseList

    file = speechDatabaseFile

    with open(file, 'rt')as f:
        data = csv.reader(f)

        for row in data:
            if row[2] != '' and row[0] != 'Set':
                if row[0] == '' and row[1] == '':
                    newPhrase = Phrase('', '', row[2])
                    phraseList.append(newPhrase)

                elif row[0] == '' and row[1] != '':
                    newPhrase = Phrase('', int(row[1]), row[2])
                    phraseList.append(newPhrase)

                elif row[0] != '' and row[1] == '':
                    newPhrase = Phrase(row[0], '', row[2])
                    phraseList.append(newPhrase)

                else:
                    newPhrase = Phrase(row[0], row[1], row[2])
                    phraseList.append(newPhrase)


# Function to check if a number is a digit including negative numbers
def _is_digit(n):
    try:
        int(n)
        return True
    except ValueError:
        return False


# Function to parse SAPI settings from voice override
# -a0 to -a100 for amplitude
# -r-10 to r10 for rate
# -v any part of the name of a SAPI voice e.g. -vHazel, -vZira
# e.g. "-a82 -r12 -vzira"
def _parseSAPIVoice(flag):
    pos = voice.find("-" + flag)
    val = ""
    if ((pos != None) and (pos >= 0)):
        val = voice[pos + 2:]
        pos = val.find(" ")
        if ((pos != None) and (pos > 0)):
            val = val[:pos]
    return val


# speak depending on synthesizer
def _speak(text):

    file = speechAudioFile
    if platform.system() == "Windows":
        if ("sapi" in synthesizer.lower()):
            from comtypes.gen import SpeechLib
            global sapivoice, sapistream

            sapistream.Open(file, SpeechLib.SSFMCreateForWrite)
            sapivoice.AudioOutputStream = sapistream

            # set any parameters
            sa = _parseSAPIVoice("a");
            if _is_digit(sa):
                sapivoice.Volume = int(sa)
            else:
                sapivoice.Volume = 100
            sr = _parseSAPIVoice("r")
            if _is_digit(sr):
                sapivoice.Rate = int(sr)
            else:
                sapivoice.Rate = 0
            sv = _parseSAPIVoice("v");
            # Default voice is always first in the list
            if (sv == ''):
                sapivoice.voice = sapivoice.GetVoices()[0]
            else:
                for v in sapivoice.GetVoices():
                    if (sv.lower() in v.GetDescription().lower()):
                        sapivoice.voice = v

            sapivoice.Speak(text)
            sapistream.Close()
        else:
            # Remove any characters that are unsafe for a subprocess call
            safetext = re.sub(r'[^ .a-zA-Z0-9?\']+', '', text)

            bashcommand = synthesizer + ' -w ' + file + ' ' + voice + ' "' + safetext + '"'
            # Execute bash command.
            subprocess.call(bashcommand, shell=True)

    if platform.system() == "Darwin":
        # Remove any characters that are unsafe for a subprocess call
        safetext = re.sub(r'[^ .a-zA-Z0-9?\']+', '', text)

        file = speechAudioFile
        bashcommand = synthesizer + file + ' --file-format=RF64 --data-format=LEI16@22050 -r' + str(
            speechRate) + ' -v ' + voice + ' "' + safetext + '"'

        # Execute bash command.
        ret = subprocess.Popen(bashcommand, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        retval = ret.wait()
        # print(retval)

    if platform.system()== "Linux":

        # Remove any characters that are unsafe for a subprocess call
        safetext = re.sub(r'[^ .a-zA-Z0-9?\']+', '', text)
        bashcommand = synthesizer + ' -w picohData/picohspeech.wav ' + voice + ' "' + safetext + '"'
        # Execute bash command.
        subprocess.call(bashcommand,shell=True)

def init(portName):
    # pickup global instances of port, ser and sapi variables   
    global port, ser, sapivoice, sapistream, connected

    _loadEyeShapes()

    dir = os.path.dirname(os.path.abspath(__file__))
    silenceFile = os.path.join(dir, 'Silence1.wav')

    # get the sapi objects ready on Windows
    if platform.system() == "Windows":
        sapivoice = CreateObject("SAPI.SpVoice")
        sapistream = CreateObject("SAPI.SpFileStream")
        winsound.PlaySound(silenceFile, winsound.SND_FILENAME)

    # get the audio system warmed up on Mac
    if platform.system() == "Darwin":
        playsound(silenceFile)

    # get the audio system warmed up on linux
    if platform.system() == "Linux":
        os.system('aplay ' + silenceFile)

    # Search for the Picoh serial port
    ports = list(serial.tools.list_ports.comports())
    for p in ports:

        # If port has Picoh connected save the location
        if portName in p[1]:
            port = p[0]
            print("Picoh found on port:" + port)
            connected = True
        elif portName in p[0]:
            port = p[0]
            print("Picoh found on port:" + port)
            connected = True

    # If not found then try the first port
    if port == "":
        for p in ports:
            port = p[0]
            print("Picoh not found")
            connected = False
            break

    if port == "":
        print("Picoh port " + portName + " not found")
        return False

    # Open the serial port
    if connected:
        ser = serial.Serial(port, 19200)
        ser.timeout = None
        ser.write_timeout = None

    # Set read timeout and write timeouts to blocking

    # Make an initial call to Festival without playing the sound to check it's all okay
    text = "Hi"

    # Create a bash command with the desired text. The command writes two files, a .wav with the speech audio and a
    # .txt file containing the phonemes and the times.
    _speak(text)
    _loadSpeechDatabase()
    return True


# Startup Code
# xml file for motor definitions
dir = os.path.dirname(os.path.abspath(__file__))
file = os.path.join(dir, 'picohdefinitions.omd')
tree = etree.parse(file)
root = tree.getroot()

# Put motor ranges into lists
motorPos = [11, 11, 11, 11, 11, 11, 11, 11]
motorMins = [0, 0, 0, 0, 0, 0, 0, 0]
motorMaxs = [0, 0, 0, 0, 0, 0, 0, 0]
motorRev = [False, False, False, False, False, False, False, False]
restPos = [0, 0, 0, 0, 0, 0, 0, 0]
isAttached = [False, False, False, False, False, False, False, False]
motorType = ["", "", "", "", "", "", "", ""]

shapesList = []

# For each line in motor defs file
for child in root:
    indexStr = child.get("Motor")
    index = int(indexStr)
    motorMins[index] = int(int(child.get("Min")) / 1000 * 180)
    motorMaxs[index] = int(int(child.get("Max")) / 1000 * 180)
    motorPos[index] = int(child.get("RestPosition"))
    restPos[index] = int(child.get("RestPosition"))
    motorType[index] = child.get("MotorType")

    if child.get("Reverse") == "True":
        rev = True
        motorRev[index] = rev
    else:
        rev = False
        motorRev[index] = rev

# initialise with any port that has USB Serial Device in the name

if platform.system() == "Windows":
    init("USB Serial Device")
if platform.system() == "Darwin":
    init("usbmodem")
if platform.system() == "Linux":
    init("Arduino")


# Function to move Picoh's motors. Arguments | m (motor) → int (0-6) | pos (position) → int (0-10) | spd (speed) →
# int (0-10) **eg move(4,3,9) or move(0,9,3)**
def move(m, pos, spd=3, eye=0):
    global lastfex, lastfey, topLipFree

    # Limit values to keep then within range
    pos = _limit(pos)
    spd = _limit(spd)

    # Keeping track of whether the top lip is pushed below the centre stop.

    if pos > 9 and m == BOTTOMLIP:
        topLipFree = True

    if pos <= 5 and m == BOTTOMLIP:
        topLipFree = False

    # Reverse the motor if necessary
    if motorRev[m]:
        pos = 10 - pos

    # Eyeturn
    if (motorType[m] == "Matrix X"):
        pos = (pos - 5) * 0.4 + 5  # Bodge to reduce the scale
        msg = "FE," + str(eye) + "," + "{:0.0f}".format(pos * 255 / 10) + "," + "{:0.0f}".format(
            lastfey * 255 / 10) + "\n"

        # print ("fex:" + msg)
        # Write message to serial port
        _serwrite(msg)
        lastfex = pos
        return

    # Eyetilt
    if (motorType[m] == "Matrix Y"):
        pos = (pos - 5) * 0.4 + 5  # Bodge to reduce the scale
        msg = "FE," + str(eye) + "," + "{:0.0f}".format(lastfex * 255 / 10) + "," + "{:0.0f}".format(
            pos * 255 / 10) + "\n"
        # print ("fey:" + msg)
        # Write message to serial port
        _serwrite(msg)
        lastfey = pos
        return

    # Blink
    if (motorType[m] == "Matrix Lid"):
        msg = "FL,+" + str(eye) + "," + "{:0.0f}".format((10 - pos) * 255 / 10) + "\n"
        # print ("fl:" + msg)
        # Write message to serial port
        _serwrite(msg)
        return

    # Attach motor       
    attach(m)

    # Convert position (0-10) to a motor position in degrees
    absPos = int(_getPos(m, pos))

    # Scale range of speed
    spd = (250 / 10) * spd

    # Construct message from values
    msg = "m0" + str(m) + "," + str(absPos) + "," + str(spd) + "\n"

    # Write message to serial port
    _serwrite(msg)

    # Update motor positions list
    motorPos[m] = pos


# Function to write to serial port
def _serwrite(s):
    global connected, writing
    if platform.system() == "Windows" or platform.system()=="Linux":
        # wait until previous write is finished
        while (writing):
            pass
        # print ('waiting on write')

    if connected:
        writing = True
        ser.write(s.encode('latin-1'))
        writing = False

    # Function to attach Picoh's motors. Argument | m (motor) int (0-6)


def attach(m):
    if isAttached[m] == False:
        # Construct message
        msg = "a0" + str(m) + "\n"

        # Write message to serial port
        _serwrite(msg)

        # Update flag
        isAttached[m] = True


# Function to detach Picoh's motors.  Argument | m (motor) int (0-6)
def detach(m):
    msg = "d0" + str(m) + "\n"
    _serwrite(msg)
    isAttached[m] = False


# Function to find the scaled position of a given motor. Arguments | m (motor) → int (0-6) | pos (position) → int (0-10) | Returns a position
def _getPos(m, pos):
    mRange = motorMaxs[m] - motorMins[m]
    scaledPos = (mRange / 10) * pos
    return scaledPos + motorMins[m]


# Function to set the voice used by the synthesiser
# name - run 'say -v ?' in terminal to find available names.
# speed - speech rate in words per min.
# This override will stay in use until it's next called
def setVoice(params=voice):
    global voice
    voice = params


# Function to set a different speech synthesizer - defaults to sapi
def setSynthesizer(params=synthesizer):
    global synthesizer
    synthesizer = params


# Set the speed of the speech in words per min.
def setSpeechSpeed(params=speechRate):
    global speechRate
    speechRate = params


# Function to make Picoh Speak. Arguments | text String "Hello World" **eg say("Hello my name is Picoh")
# untilDone - wait in function until speech is complete, lipSync - move lips in time with speech, hdmiAudio - adds a delay to give hdmi channel time to activate.
# soundDelay - positive if lip movement is lagging behind sound, negative if sound is lagging behind lip movement.
def say(text, untilDone=True, lipSync=True, hdmiAudio=False, soundDelay=0):
    global topLipFree

    if text.isspace() or text == '':
        return

    # Use the bottom lip to push the top lip above the centre point if it is below

    if topLipFree:
        move(BOTTOMLIP, 4)
        wait(0.25)

    text = text.replace("picoh", "peek oh")
    text = text.replace("Picoh", "peek oh")

    if platform.system() == "Linux":
        _sayLinux(text,untilDone,lipSync,hdmiAudio,soundDelay)
        return

    if hdmiAudio:
        soundDelay = soundDelay - 1



    # Create a bash command with the desired text. espeak.exe or another synthesizer must be in the current folder.  the -w parameter forces the speech to a file
    _speak(text)

    # open the file to calculate visemes. Festival on RPi has this built in but for espeak need to do it manually

    dir = os.path.dirname(os.path.abspath(__file__))
    file = speechAudioFile

    waveFile = wave.open(file, 'r')

    length = waveFile.getnframes()
    framerate = waveFile.getframerate()
    channels = waveFile.getnchannels()
    bytespersample = waveFile.getsampwidth()

    # How many samples per second for mouth position
    if platform.system() == "Windows":
        VISEMESPERSEC = 10
    if platform.system() == "Darwin":
        VISEMESPERSEC = 10
    if platform.system() == "Linux":
        VISEMESPERSEC = 10

    # How many samples in 1/20th second
    # print ('framerate:', framerate, ' channels:', channels, ' length:', length, ' bytespersample:', bytespersample)

    chunk = int(waveFile.getframerate() / VISEMESPERSEC)
    # print ('chunk:', chunk)

    # Empty the lists that contain phoneme data and reset count
    phonemes = []
    times = []

    ms = 0

    for i in range(0, length - chunk, chunk):
        vol = 0
        buffer = waveFile.readframes(chunk)
        # frame is 1 sample for mono or 2 for stereo
        bytesread = chunk * channels * bytespersample
        # print ('bytesread:', bytesread)
        index = 0;
        for sample in range(0, int(bytesread / (channels * bytespersample))):
            vol += buffer[index]
            vol += buffer[index + 1] * 256
            index += bytespersample
            if channels > 1:
                vol += buffer[index]
                vol += buffer[index + 1] * 256
                index += bytespersample

        # print ('viseme', i, ":", ms, ':', vol)
        ms += (1000 / VISEMESPERSEC);

        phonemes.append(float(vol))
        times.append(float(ms) / 1000)

    # Back to the beginning for next use
    waveFile.rewind()

    # Normalise the volume
    max = 0
    for i in range(0, len(phonemes) - 1):
        if phonemes[i] > max:
            max = phonemes[i]

    for i in range(0, len(phonemes) - 1):
        phonemes[i] = phonemes[i] * 10 / max
        # print ('visnorm', i, ":", times[i], ':', phonemes[i])

    if lipSync:
        if soundDelay > 0:
            # Set up a thread for the speech sound synthesis, delay start by soundDelay
            t = threading.Timer(soundDelay, _saySpeech, args=(hdmiAudio,), kwargs=None)
            # Set up a thread for the speech movement
            t2 = threading.Thread(target=_moveSpeech, args=(phonemes, times))
        else:
            # Set up a thread for the speech sound synthesis
            t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
            # Set up a thread for the speech movement, delay start by - soundDelay
            t2 = threading.Timer(-soundDelay, _moveSpeech, args=(phonemes, times), kwargs=None)
        t2.start()
    else:
        # Set up a thread for the speech sound synthesis
        t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
    t.start()

    # if untilDone, keep running until speech has finished    
    if untilDone:
        totalTime = times[len(times) - 1]
        startTime = time.time()
        while time.time() - startTime < totalTime:
            continue


def _sayLinux(text, untilDone=True, lipSync=True, hdmiAudio=False, soundDelay=0):


    if "festival" in synthesizer.lower():

        if hdmiAudio:
            soundDelay = soundDelay - 1

        safetext = re.sub(r'[^ .a-zA-Z0-9?\']+', '', text)

        # Create a bash command with the desired text. The command writes two files, a .wav with the speech audio and a .txt file containing the phonemes and the times.

        bashcommand = "festival -b '(set! mytext (Utterance Text " + '"' + safetext + '"))' + "' '(utt.synth mytext)' '(utt.save.wave mytext " + '"picohData/picohspeech.wav")' + "' '(utt.save.segs mytext " + '"picohData/phonemes"' + ")'"

        # Execute bash command.
        subprocess.call(bashcommand, shell=True)

        # Open the text file containing the phonemes

        f = open("picohData/phonemes", 'r')

        # Empty the lists that contain phoneme data and reset count
        phonemes = []
        times = []
        vals = []

        # Read a line to move past the first line
        line = f.readline()

        # While there are more lines to read.
        while line:

            # Read the line
            line = f.readline()

            # Split the line into values
            vals = line.split()

            # If values exist add the phoneme to the phonemes list and the timecode to the times list.
            if vals:
                phonemes.append(vals[2])
                times.append(float(vals[0]))

        if lipSync:
            if soundDelay > 0:
                # Set up a thread for the speech sound synthesis, delay start by soundDelay
                t = threading.Timer(soundDelay, _saySpeech, args=(hdmiAudio,), kwargs=None)
                # Set up a thread for the speech movement
                t2 = threading.Thread(target=_moveSpeechFest, args=(phonemes, times))
            else:
                # Set up a thread for the speech sound synthesis
                t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
                # Set up a thread for the speech movement, delay start by - soundDelay
                t2 = threading.Timer(-soundDelay, _moveSpeechFest, args=(phonemes, times), kwargs=None)
            t2.start()
        else:
            # Set up a thread for the speech sound synthesis
            t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
        t.start()

        # if untilDone, keep running until speech has finished
        if untilDone:
            totalTime = times[len(times) - 1]
            startTime = time.time()
            while time.time() - startTime < totalTime:
                continue

    if ("espeak" in synthesizer.lower() or "pico2wave" in synthesizer.lower()):

        if hdmiAudio:
            soundDelay = soundDelay - 1

        # Create a bash command with the desired text. espeak.exe or another synthesizer must be in the current folder.  the -w parameter forces the speech to a file
        _speak(text)

        # open the file to calculate visemes. Festival on RPi has this built in but for espeak need to do it manually
        waveFile = wave.open("picohData/picohspeech.wav", 'r')

        length = waveFile.getnframes()
        framerate = waveFile.getframerate()
        channels = waveFile.getnchannels()
        bytespersample = waveFile.getsampwidth()

        # How many samples per second for mouth position
        VISEMESPERSEC = 20

        # How many samples in 1/20th second
        # print ('framerate:', framerate, ' channels:', channels, ' length:', length, ' bytespersample:', bytespersample)

        chunk = int(waveFile.getframerate() / VISEMESPERSEC)
        # print ('chunk:', chunk)

        # Empty the lists that contain phoneme data and reset count
        phonemes = []
        times = []

        ms = 0

        for i in range(0, length - chunk, chunk):
            vol = 0
            buffer = waveFile.readframes(chunk)
            # frame is 1 sample for mono or 2 for stereo
            bytesread = chunk * channels * bytespersample
            # print ('bytesread:', bytesread)
            index = 0;
            for sample in range(0, int(bytesread / (channels * bytespersample))):
                vol += buffer[index]
                vol += buffer[index + 1] * 256
                index += bytespersample
                if channels > 1:
                    vol += buffer[index]
                    vol += buffer[index + 1] * 256
                    index += bytespersample

            # print ('viseme', i, ":", ms, ':', vol)
            ms += (1000 / VISEMESPERSEC);

            phonemes.append(float(vol))
            times.append(float(ms) / 1000)

        # Back to the beginning for next use
        waveFile.rewind()

        # Normalise the volume
        max = 0
        for i in range(0, len(phonemes) - 1):
            if (phonemes[i] > max):
                max = phonemes[i]

        for i in range(0, len(phonemes) - 1):
            phonemes[i] = phonemes[i] * 10 / max
            # print ('visnorm', i, ":", times[i], ':', phonemes[i])

        if lipSync:
            if soundDelay > 0:
                # Set up a thread for the speech sound synthesis, delay start by soundDelay
                t = threading.Timer(soundDelay, _saySpeech, args=(hdmiAudio,), kwargs=None)
                # Set up a thread for the speech movement
                t2 = threading.Thread(target=_moveSpeech, args=(phonemes, times))
            else:
                # Set up a thread for the speech sound synthesis
                t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
                # Set up a thread for the speech movement, delay start by - soundDelay
                t2 = threading.Timer(-soundDelay, _moveSpeech, args=(phonemes, times), kwargs=None)
            t2.start()
        else:
            # Set up a thread for the speech sound synthesis
            t = threading.Thread(target=_saySpeech, args=(hdmiAudio,))
        t.start()

        # if untilDone, keep running until speech has finished
        if untilDone:
            totalTime = times[len(times) - 1]
            startTime = time.time()
            while time.time() - startTime < totalTime:
                continue


# Function to limit values so they are between 0 - 10
def _limit(val):
    if val > 10:
        return 10
    elif val < 0:
        return 0
    else:
        return val


# Function to play back the speech wav file, if hmdi audio is being used play silence before speech sound
def _saySpeech(addSilence):
    dir = os.path.dirname(os.path.abspath(__file__))
    #speechFile = os.path.join(dir, 'picohspeech.wav')
    silenceFile = os.path.join(dir, 'Silence1.wav')

    speechFile = speechAudioFile
    if platform.system() == "Windows":

        if addSilence:
            winsound.PlaySound(silenceFile, winsound.SND_FILENAME)
        winsound.PlaySound(speechFile, winsound.SND_FILENAME)

    if platform.system() == "Darwin":
        if addSilence:
            playsound(silenceFile)
        playsound(speechFile)

    if platform.system() == "Linux":
        if addSilence:
            commandString = 'aplay ' + silenceFile + '\naplay picohData/picohspeech.wav'
            os.system(commandString)
        else:
            os.system('aplay picohData/picohspeech.wav')


# Function to move Picoh's lips in time with speech. Arguments | phonemes → list of phonemes[] | waits → list of waits[]
def _moveSpeech(phonemes, times):
    startTime = time.time()
    timeNow = 0
    totalTime = times[len(times) - 1]
    currentX = -1
    while timeNow < totalTime:
        timeNow = time.time() - startTime
        for x in range(0, len(times)):
            if timeNow > times[x] and x > currentX:
                posTop = _phonememapTop(phonemes[x])
                posBottom = _phonememapBottom(phonemes[x])
                move(TOPLIP, posTop, 10)
                move(BOTTOMLIP, posBottom, 10)
                currentX = x
    move(TOPLIP, 5)
    move(BOTTOMLIP, 5)


def _moveSpeechFest(phonemes, times):
    startTime = time.time()
    timeNow = 0
    totalTime = times[len(times)-1]
    currentX = -1
    while timeNow < totalTime:
        timeNow = time.time() - startTime
        for x in range (0,len(times)):
            if timeNow > times[x] and x > currentX:
                posTop = _phonememapTopFest(phonemes[x])
                posBottom = _phonememapBottomFest(phonemes[x])
                move(TOPLIP,posTop,10)
                move(BOTTOMLIP,posBottom,10)
                currentX = x
    move(TOPLIP,5)
    move(BOTTOMLIP,5)


# Function mapping phonemes to top lip positions. Argument | val → phoneme | returns a position as int
def _phonememapTopFest(val):
    return {
        'p': 5,
        'b': 5,
        'm': 5,
        'ae': 7,
        'ax': 7,
        'ah': 7,
        'aw': 10,
        'aa': 10,
        'ao': 10,
        'ow': 10,
        'ey': 7,
        'eh': 7,
        'uh': 7,
        'ay': 7,
        'h': 7,
        'er': 8,
        'r': 8,
        'l': 8,
        'y': 6,
        'iy': 6,
        'ih': 6,
        'ix': 6,
        'w': 6,
        'uw': 6,
        'oy': 6,
        's': 5,
        'z': 5,
        'sh': 5,
        'ch': 5,
        'jh': 5,
        'zh': 5,
        'th': 5,
        'dh': 5,
        'd': 5,
        't': 5,
        'n': 5,
        'k': 5,
        'g': 5,
        'ng': 5,
        'f': 6,
        'v': 6
    }.get(val, 5)


# Function mapping phonemes to lip positions. Argument | val → phoneme | returns a position as int
def _phonememapBottomFest(val):
    return {
        'p': 5,
        'b': 5,
        'm': 5,
        'ae': 8,
        'ax': 8,
        'ah': 8,
        'aw': 5,
        'aa': 10,
        'ao': 10,
        'ow': 10,
        'ey': 7,
        'eh': 7,
        'uh': 7,
        'ay': 7,
        'h': 7,
        'er': 8,
        'r': 8,
        'l': 8,
        'y': 6,
        'iy': 6,
        'ih': 6,
        'ix': 6,
        'w': 6,
        'uw': 6,
        'oy': 6,
        's': 6,
        'z': 6,
        'sh': 6,
        'ch': 6,
        'jh': 6,
        'zh': 6,
        'th': 6,
        'dh': 6,
        'd': 6,
        't': 6,
        'n': 6,
        'k': 6,
        'g': 6,
        'ng': 6,
        'f': 5,
        'v': 5
    }.get(val, 5)


# Function mapping phonemes to top lip positions.
def _phonememapTop(val):
    return 5 + (_limit(val) / 2)


# Function mapping phonemes to top lip positions.
# Bottom lip never goes over 9
def _phonememapBottom(val):
    return 5 + (_limit(val) * 4 / 10)


# Legacy function to support Ohbot programs with eyeColour. Passes onto baseColour.
def eyeColour(r, g, b, swapRandG=False):
    baseColour(r, g, b, swapRandG)


# Function to set the color of the LEDs in Picoh's base. Arguments | r (red) → int (0-10) | g (green) → int (0-10) | b (blue) → int (0-10)
# swapRandG is used to swap the red and green values as this is required for some LEDs
def baseColour(r, g, b, swapRandG=False):
    # Limit the values to keep them within range.
    r = _limit(r)
    g = _limit(g)
    b = _limit(b)

    # Scale the values so they are between 0 - 255.
    r = int((255 / 10) * r)

    g = int((255 / 10) * g)

    b = int((255 / 10) * b)

    # Construct a message with the values.
    if swapRandG:
        msg1 = "l00," + str(g) + "," + str(r) + "," + str(b) + "\n"
        msg2 = "l01," + str(g) + "," + str(r) + "," + str(b) + "\n"
    else:
        msg1 = "l00," + str(r) + "," + str(g) + "," + str(b) + "\n"
        msg2 = "l01," + str(r) + "," + str(g) + "," + str(b) + "\n"

    # Write message to serial port.
    _serwrite(msg1)
    _serwrite(msg2)


def wait(seconds):
    time.sleep(float(seconds))
    return


def close():
    for x in range(0, len(motorMins) - 1):
        detach(x)

    # Reset Picoh back to start position


def reset():
    baseColour(0, 0, 0)
    for x in range(0, len(restPos) - 1):
        move(x, restPos[x])


# Return the sensor value between 0-10 for a given sensor number. Values stored in sensors[] array.
def readSensor(index):
    ser.flushInput()

    msg = "i0" + str(index) + "\n"
    _serwrite(msg)

    line = ser.readline()
    lineStr = line.decode("utf-8")
    lines = lineStr.split(",")

    if len(lines) > 1:
        indexIn = lines[0]
        indexIn = indexIn[1]

        intdex = int(indexIn)

        newVal = int(lines[1]) / 1024
        newVal = newVal * 10
        sensors[intdex] = _limit(newVal)

    return sensors[index]


# set the brightness of the eyes.  Value is 0 (off) to 10 (full brightness)
def setEyeBrightness(val):
    val = val / 10
    val = val * val
    msg = "{:0.0f}".format(val * 255)

    # print ("brightness:" + msg)

    _serwrite("FI," + msg + "\n")


# set the eye shape according to the passed in eyeshapedefinition
# eysshape definition is 6 sets of 9 hex pairs which set the bits of half of the screen
# the other half of the screen is a mirror copy
# the first 5 sets of pairs set the normal eye and 4 blink positions
# the last set of pairs set the pupil 
# set the eye shape according to the passed in eyeshapedefinition
# eyeshape definition is 6 sets of 9 hex pairs which set the bits of half of the screen
# the other half of the screen is a mirror copy
# the first 5 sets of pairs set the normal eye and 4 blink positions
# the last set of pairs set the pupil
def _setEyes(leftDefinition, rightDefinition="", autoMirror=True):
    # TODO needs some work here to use both definitions for independent eye shapes
    # and to define the pupil range and offset
    definition = leftDefinition

    if rightDefinition == "":
        rightDefinition = definition

    _serwrite("FB,0," + _EyeShapeBytes(definition, rightDefinition, 0, autoMirror) + "\n")
    _serwrite("FB,1," + _EyeShapeBytes(definition, rightDefinition, 1, autoMirror) + "\n")
    _serwrite("FB,2," + _EyeShapeBytes(definition, rightDefinition, 2, autoMirror) + "\n")
    _serwrite("FB,3," + _EyeShapeBytes(definition, rightDefinition, 3, autoMirror) + "\n")
    _serwrite("FB,4," + _EyeShapeBytes(definition, rightDefinition, 4, autoMirror) + "\n")
    # Pupil is held in set 6 and has been implemented in the Arduino driver as FB 8
    _serwrite("FB,8," + _EyeShapeBytes(definition, rightDefinition, 5, autoMirror) + "\n")


# function for getting a string that defines the 16 x 9 matrix for a particular 8 x 9 eyeshapedefinition
# setNo is 0 for the normal eyeshape, 1 to 4 for blink shapes or 5 for the pupil
def _EyeShapeBytes(definitionR, definition, setNo, autoMirror):
    strRet = ""
    for x in range(0, 9):
        if (len(strRet) > 0):
            strRet += ","
        offset = setNo * 18 + x * 2

        if autoMirror:
            strRet += (definition[offset: offset + 2])
        else:
            strRet += _reverseBits(definition[offset: offset + 2])

        strRet += _reverseBits(definitionR[offset: offset + 2])

    # print ("eyeshape set:" + str(set) + ": " + strRet)

    return strRet


# reverse the bits of a two byte hex number
def _reverseBits(str):
    # print ("str: " + str)
    x = int(str, 16)
    r = 0

    if (x & 0x80):
        r += 0x01
    if (x & 0x40):
        r += 0x02
    if (x & 0x20):
        r += 0x04
    if (x & 0x10):
        r += 0x08
    if (x & 0x08):
        r += 0x10
    if (x & 0x04):
        r += 0x20
    if (x & 0x02):
        r += 0x40
    if (x & 0x01):
        r += 0x80

    # https://stackoverflow.com/questions/2269827/how-to-convert-an-int-to-a-hex-string
    return "%0.2X" % r


def setEyeShape(shapeNameLeft, shapeNameRight=''):
    global shapeList

    if shapeNameRight == '':
        shapeNameRight = shapeNameLeft

    leftHex = ''

    for index, shape in enumerate(shapeList):
        if shape.name == shapeNameLeft:
            leftHex = shape.hexString
            if shape.autoMirror:
                autoMirrorVar = True
            else:
                autoMirrorVar = False

    for index, shape in enumerate(shapeList):
        if shape.name == shapeNameRight:
            rightHex = shape.hexString
            if shape.autoMirror:
                autoMirrorVar = True
            else:
                autoMirrorVar = False
    #    print(autoMirrorVar)
    # Send hex to Picoh.

    if leftHex == '':
        print(str(shapeNameLeft) + " Eyeshape Not Found")
        return
    if connected:
        _setEyes(rightHex, leftHex, autoMirrorVar)
        wait(0.05)
        move(EYETILT, motorPos[EYETILT])
        move(EYETURN, motorPos[EYETURN])


def getPhrase(set='None', variable='None'):
    global phraseList
    possiblePhrases = []
    for phrase in phraseList:
        if str(set) == str(phrase.set) or set == 'None':
            if str(variable) == phrase.variable or variable == 'None':
                possiblePhrases.append(phrase.text)
    length = len(possiblePhrases)
    if length == 0:
        print(
            "No matching phrase found for set: " + str(set) + " variable: " + str(variable) + " in: " + speechFile)
        return ""
    elif length == 1:
        return possiblePhrases[0]
    else:
        return possiblePhrases[random.randint(0, length - 1)]
